﻿
' <auto-generated />

Imports System.Linq
Imports System.Text
Imports System.Collections
Imports System.Reflection
Imports System.Collections.Generic
Imports Microsoft.Xrm.Sdk
Imports Microsoft.Xrm.Sdk.Query
Imports Microsoft.Crm.Sdk.Messages
Imports Microsoft.Xrm.Sdk.Messages
Imports Microsoft.Xrm.Sdk.Metadata
Namespace ProxyClassNameSpace
    Partial Public MustInherit Class BaseProxyClass
        Protected Sub New(Original As Entity)
            If String.IsNullOrEmpty(Original.LogicalName) Then
                Throw New Exception("Please specify the 'LogicalName' on the entity when using a proxy class.")
            End If
            Dim EntityName As String = GetLogicalName(Me.[GetType]())
            If EntityName <> Original.LogicalName Then
                Throw New Exception("Please make sure that the entity logical name matches that of the proxy class you are creating.")
            End If
            _entity = Original
        End Sub
        Public Shared ReturnDatesInLocalTime As Boolean = False
        Private ReadOnly _entity As Entity
        Public ReadOnly Property Entity() As Entity
            Get
                Return _entity
            End Get
        End Property
        Public Property Id() As Guid
            Get
                Return Me.Entity.Id
            End Get
            Set(value As Guid)
                Me.Entity.Id = value
            End Set
        End Property
        Public Property LogicalName() As String
            Get
                Return Me.Entity.LogicalName
            End Get
            Set(value As String)
                Me.Entity.LogicalName = value
            End Set
        End Property
        Public Function Create(Service As IOrganizationService) As Guid
            Me.Id = Service.Create(Me.Entity)
            mChangedValues.Clear()
            Return Me.Id
        End Function
        ''' <summary>
        ''' Checks to see if the underlying entity contains the specified attribute
        ''' </summary>
        ''' <param name="AttributeName">Name of attribute to check</param>
        ''' <returns></returns>
        Public Function Contains(AttributeName As String) As Boolean
            Return Me.Entity.Contains(AttributeName)
        End Function
        Public Sub Delete(Service As IOrganizationService)
            Service.Delete(Me.Entity.LogicalName, Me.Entity.Id)
        End Sub
        Public Sub Update(Service As IOrganizationService)
            If mChangedValues.Count > 0 Then
                Service.Update(GetChangedEntity())
            End If
            mChangedValues.Clear()
        End Sub
        Public Function GetChangedEntity() As Entity
            Dim entity As New Entity(Me.LogicalName)
            entity.Id = Me.Id
            For Each attributeName As String In mChangedValues.Keys
                entity.Attributes(attributeName) = Me.Entity.Attributes(attributeName)
            Next
            Return entity
        End Function
        Public Sub Save(Service As IOrganizationService)
            If Me.Entity.Id <> Guid.Empty Then
                Me.Update(Service)
            Else
                Me.Create(Service)
            End If
        End Sub
        Public Shared Function CreateProxyType(entity As Entity) As BaseProxyClass
            If ProxyTypes.ContainsValue(entity.LogicalName) Then
                Dim ProxyType As Type = (From ProxyTypeTemp In ProxyTypes.Keys Where ProxyTypes(ProxyTypeTemp) = entity.LogicalName Select ProxyTypeTemp).FirstOrDefault()
                Dim ConstructorTemp As ConstructorInfo = ProxyType.GetConstructor(New Type() {GetType(Entity)})
                Return DirectCast(ConstructorTemp.Invoke(New Object() {entity}), BaseProxyClass)
            Else
                Throw New Exception("The proxy class for type '" + entity.LogicalName + "' has not been registered.")
            End If
        End Function
        Private Shared ProxyTypes As New Dictionary(Of Type, String)()
        Public Shared Sub RegsterProxyType(ProxyType As Type, LogicalName As String)
            If Not ProxyTypes.ContainsKey(ProxyType) Then
                ProxyTypes.Add(ProxyType, LogicalName)
            End If
        End Sub
        Public Shared Function GetLogicalName(Of T As BaseProxyClass)() As String
            Return GetLogicalName(GetType(T))
        End Function
        Public Shared Function GetLogicalName(type As Type) As String
            If Not ProxyTypes.ContainsKey(type) Then
                'This forces the static constructor to be called and register the name
                Dim InfoTemp As ConstructorInfo = type.GetConstructor(New Type() {})
                InfoTemp.Invoke(New Object() {})
            End If
            Return ProxyTypes(type)
        End Function
        Public Function ToEntityReference() As EntityReference
            If Me.Id <> Guid.Empty Then
                Return New EntityReference(Me.LogicalName, Me.Id)
            End If
            Return Nothing
        End Function
        Public Shared Widening Operator CType(Proxy As BaseProxyClass) As EntityReference
            If Proxy IsNot Nothing Then
                Return Proxy.ToEntityReference()
            End If
            Return Nothing
        End Operator
        Public Shared Widening Operator CType(Proxy As BaseProxyClass) As Entity
            If Proxy IsNot Nothing Then
                Return Proxy.Entity
            End If
            Return Nothing
        End Operator
        Public Function GetPropertyValue(Of T)(Name As String) As T
            If Me.Entity.Contains(Name) Then
                Dim Value As T = DirectCast(Me.Entity.Attributes(Name), T)
                If (GetType(T) = GetType(DateTime) OrElse GetType(T) = GetType(System.Nullable(Of DateTime))) AndAlso ReturnDatesInLocalTime AndAlso Value IsNot Nothing Then
                    Value = DirectCast(DirectCast(DirectCast(DirectCast(Value, [Object]), DateTime).ToLocalTime(), [Object]), T)
                End If
                Return Value
            End If
            Return Nothing
        End Function
        'dictionary of values that have changed along with the original value
        Private mChangedValues As New Dictionary(Of String, Object)()
        Private mEqualityComparer As New AttributeEqualityComparer()
        Public Sub SetPropertyValue(Of T)(Name As String, Value As T)
            If mChangedValues.ContainsKey(Name) Then
                Dim OriginalValue As Object = mChangedValues(Name)
                Dim CurrentValue As Object = If(Me.Entity.Contains(Name), DirectCast(Me.GetPropertyValue(Of T)(Name), Object), Nothing)
                If Not mEqualityComparer.Equals(CurrentValue, Value) Then
                    If mEqualityComparer.Equals(OriginalValue, Value) Then
                        mChangedValues.Remove(Name)
                    End If
                End If
            Else
                Dim CurrentValue As Object = If(Me.Entity.Contains(Name), DirectCast(Me.GetPropertyValue(Of T)(Name), Object), Nothing)
                If Not mEqualityComparer.Equals(CurrentValue, Value) Then
                    mChangedValues.Add(Name, CurrentValue)
                End If
            End If
            Me.Entity.Attributes(Name) = Value
        End Sub
        Public Sub SetPropertyValue(Name As String, Value As String, MaxLength As Integer)
            Dim TextOptions As eTextOptions = GetTextOptions(Name)
            If TextOptions <> eTextOptions.Ignore AndAlso Not String.IsNullOrEmpty(Value) AndAlso Value.Length > MaxLength Then
                If TextOptions = eTextOptions.Truncate Then
                    Value = Value.Substring(0, MaxLength)
                Else
                    Throw New Exception(String.Format(GetErrorString(Name, eErrorType.Text), Name, Value, Value.Length, MaxLength))
                End If
            End If
            SetPropertyValue(Of String)(Name, Value)
        End Sub
        Public Sub SetPropertyValue(Name As String, Value As System.Nullable(Of Integer), MinValue As Integer, MaxValue As Integer)
            Dim NumberOptions As eNumberOptions = GetNumberOptions(Name)
            If NumberOptions <> eNumberOptions.Ignore AndAlso (Value < MinValue OrElse Value > MaxValue) Then
                Dim ThrowError As Boolean = False
                If NumberOptions = eNumberOptions.CorrectMinAndMax Then
                    Value = If((Value < MinValue), MinValue, MaxValue)
                ElseIf NumberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    Value = If((Value < MinValue), MinValue, Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMinThrowMax AndAlso Value < MinValue Then
                    Value = MinValue
                ElseIf NumberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    Value = If((Value > MaxValue), MaxValue, Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso Value > MaxValue Then
                    Value = MaxValue
                Else
                    ThrowError = True
                End If
                If ThrowError Then
                    Throw New Exception(String.Format(GetErrorString(Name, eErrorType.Number), Name, Value, MinValue, MaxValue))
                End If
            End If
            SetPropertyValue(Of System.Nullable(Of Integer))(Name, Value)
        End Sub
        Public Sub SetPropertyValue(Name As String, Value As System.Nullable(Of Decimal), MinValue As Decimal, MaxValue As Decimal)
            Dim NumberOptions As eNumberOptions = GetNumberOptions(Name)
            If NumberOptions <> eNumberOptions.Ignore AndAlso (Value < MinValue OrElse Value > MaxValue) Then
                Dim ThrowError As Boolean = False
                If NumberOptions = eNumberOptions.CorrectMinAndMax Then
                    Value = If((Value < MinValue), MinValue, MaxValue)
                ElseIf NumberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    Value = If((Value < MinValue), MinValue, Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMinThrowMax AndAlso Value < MinValue Then
                    Value = MinValue
                ElseIf NumberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    Value = If((Value > MaxValue), MaxValue, Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso Value > MaxValue Then
                    Value = MaxValue
                Else
                    ThrowError = True
                End If
                If ThrowError Then
                    Throw New Exception(String.Format(GetErrorString(Name, eErrorType.Number), Name, Value, MinValue, MaxValue))
                End If
            End If
            SetPropertyValue(Of System.Nullable(Of Decimal))(Name, Value)
        End Sub
        Public Sub SetPropertyValue(Name As String, Value As System.Nullable(Of Double), MinValue As Double, MaxValue As Double)
            Dim NumberOptions As eNumberOptions = GetNumberOptions(Name)
            If NumberOptions <> eNumberOptions.Ignore AndAlso (Value < MinValue OrElse Value > MaxValue) Then
                Dim ThrowError As Boolean = False
                If NumberOptions = eNumberOptions.CorrectMinAndMax Then
                    Value = If((Value < MinValue), MinValue, MaxValue)
                ElseIf NumberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    Value = If((Value < MinValue), MinValue, Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMinThrowMax AndAlso Value < MinValue Then
                    Value = MinValue
                ElseIf NumberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    Value = If((Value > MaxValue), MaxValue, Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso Value > MaxValue Then
                    Value = MaxValue
                Else
                    ThrowError = True
                End If
                If ThrowError Then
                    Throw New Exception(String.Format(GetErrorString(Name, eErrorType.Number), Name, Value, MinValue, MaxValue))
                End If
            End If
            SetPropertyValue(Of System.Nullable(Of Double))(Name, Value)
        End Sub
        Public Sub SetPropertyValue(Name As String, Value As Money, MinValue As Decimal, MaxValue As Decimal)
            Dim NumberOptions As eNumberOptions = GetNumberOptions(Name)
            If Value IsNot Nothing AndAlso NumberOptions <> eNumberOptions.Ignore AndAlso (Value.Value < MinValue OrElse Value.Value > MaxValue) Then
                Dim ThrowError As Boolean = False
                If NumberOptions = eNumberOptions.CorrectMinAndMax Then
                    Value.Value = If((Value.Value < MinValue), MinValue, MaxValue)
                ElseIf NumberOptions = eNumberOptions.CorrectMinIgnoreMax Then
                    Value.Value = If((Value.Value < MinValue), MinValue, Value.Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMinThrowMax AndAlso Value.Value < MinValue Then
                    Value.Value = MinValue
                ElseIf NumberOptions = eNumberOptions.CorrectMaxIgnoreMin Then
                    Value.Value = If((Value.Value > MaxValue), MaxValue, Value.Value)
                ElseIf NumberOptions = eNumberOptions.CorrectMaxThrowMin AndAlso Value.Value > MaxValue Then
                    Value.Value = MaxValue
                Else
                    ThrowError = True
                End If
                If ThrowError Then
                    Throw New Exception(String.Format(GetErrorString(Name, eErrorType.Number), Name, Value.Value, MinValue, MaxValue))
                End If
            End If
            SetPropertyValue(Of Money)(Name, Value)
        End Sub
        Protected MustOverride Function GetTextOptions(LogicalName As String) As eTextOptions
        Protected MustOverride Function GetErrorString(AttributeName As String, DefaultErrorType As eErrorType) As String
        Protected MustOverride Function GetNumberOptions(LogicalName As String) As eNumberOptions
        Public ReadOnly Property IsDirty() As Boolean
            Get
                Return Me.mChangedValues.Count > 0
            End Get
        End Property
        Public Shared Function GetRelatedOneToManyEntities(Of T As BaseProxyClass)(Service As IOrganizationService, RelatedId As Guid, LogicalName As String, AttributeName As String, ParamArray Columns As String()) As List(Of T)
            Return GetRelatedOneToManyEntities(Service, RelatedId, LogicalName, AttributeName, Columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedOneToManyEntities(Of T As BaseProxyClass)(Service As IOrganizationService, RelatedId As Guid, LogicalName As String, AttributeName As String, Columns As ColumnSet) As List(Of T)
            Return GetRelatedOneToManyEntities(Service, RelatedId, LogicalName, AttributeName, Columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedOneToManyEntities(Service As IOrganizationService, RelatedId As Guid, LogicalName As String, AttributeName As String, ParamArray Columns As String()) As List(Of Entity)
            Dim ColumnSet As ColumnSet = If((Columns.Length = 0), New ColumnSet(True), New ColumnSet(Columns))
            Return GetRelatedOneToManyEntities(Service, RelatedId, LogicalName, AttributeName, ColumnSet)
        End Function
        Public Shared Function GetRelatedOneToManyEntities(Service As IOrganizationService, RelatedId As Guid, LogicalName As String, AttributeName As String, Columns As ColumnSet) As List(Of Entity)
            Return Service.RetrieveMultiple(LogicalName, New ConditionExpression(AttributeName, ConditionOperator.Equal, RelatedId), Columns).Entities.ToList()
        End Function
        Public Shared Function GetRelatedManyToManyEntities(Of T As BaseProxyClass)(service As IOrganizationService, proxy As BaseProxyClass, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String, intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, ParamArray columns As String()) As List(Of T)
            Return GetRelatedManyToManyEntities(service, proxy, linkFromEntityTypeName, linkFromAttributeName, linkToEntityTypeName, linkToAttributeName,
        intermediateEntityName, intermediateFromAttributeName, intermediateToAttributeName, columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedManyToManyEntities(Of T As BaseProxyClass)(service As IOrganizationService, proxy As BaseProxyClass, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String, intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, columns As ColumnSet) As List(Of T)
            Return GetRelatedManyToManyEntities(service, proxy, linkFromEntityTypeName, linkFromAttributeName, linkToEntityTypeName, linkToAttributeName,
        intermediateEntityName, intermediateFromAttributeName, intermediateToAttributeName, columns).ToProxies(Of T)()
        End Function
        Public Shared Function GetRelatedManyToManyEntities(service As IOrganizationService, proxy As BaseProxyClass, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String, intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, ParamArray columns As String()) As List(Of Entity)
            Dim columnSet = If((columns.Length = 0), New ColumnSet(True), New ColumnSet(columns))
            Return GetRelatedManyToManyEntities(service, proxy, linkFromEntityTypeName, linkFromAttributeName, linkToEntityTypeName, linkToAttributeName,
        intermediateEntityName, intermediateFromAttributeName, intermediateToAttributeName, columnSet)
        End Function
        Public Shared Function GetRelatedManyToManyEntities(service As IOrganizationService, proxy As BaseProxyClass, linkFromEntityTypeName As String, linkFromAttributeName As String, linkToEntityTypeName As String, linkToAttributeName As String, intermediateEntityName As String, intermediateFromAttributeName As String, intermediateToAttributeName As String, columns As ColumnSet) As List(Of Entity)
            Dim query = New QueryExpression() With {
                .EntityName = linkToEntityTypeName,
                .ColumnSet = Columns
            }

            query.LinkEntities.Add(New LinkEntity() With {
                .LinkFromAttributeName = linkToAttributeName,
                .LinkFromEntityName = linkToEntityTypeName,
                .LinkToAttributeName = intermediateToAttributeName,
                .LinkToEntityName = intermediateEntityName})

            query.LinkEntities(0).LinkEntities.Add(New LinkEntity() With {
                    .LinkFromAttributeName = intermediateFromAttributeName,
                    .LinkFromEntityName = intermediateEntityName,
                    .LinkToAttributeName = linkFromAttributeName,
                    .LinkToEntityName = linkFromEntityTypeName})

            query.LinkEntities(0).LinkEntities(0).LinkCriteria.Conditions.Add(New ConditionExpression(linkFromAttributeName, ConditionOperator.Equal, New Object() {proxy.Id}))

            Return service.RetrieveMultiple(query).Entities.ToList()
        End Function
        Private Class AttributeEqualityComparer
            Implements IEqualityComparer
            Public Overloads Function Equals(x As Object, y As Object) As Boolean Implements IEqualityComparer.Equals
                If (x Is Nothing OrElse (x.[GetType]() Is GetType(String) AndAlso String.IsNullOrEmpty(TryCast(x, String)))) AndAlso (y Is Nothing OrElse (y.[GetType]() Is GetType(String) AndAlso String.IsNullOrEmpty(TryCast(y, String)))) Then
                    Return True
                Else
                    If x Is Nothing AndAlso y Is Nothing Then
                        Return True
                    ElseIf x Is Nothing AndAlso y IsNot Nothing Then
                        Return False
                    ElseIf x IsNot Nothing AndAlso y Is Nothing Then
                        Return False
                    ElseIf x.GetType() Is y.GetType() Then
                        If x.GetType() Is GetType(OptionSetValue) Then
                            Return DirectCast(x, OptionSetValue).Value = DirectCast(y, OptionSetValue).Value
                        ElseIf x.GetType() Is GetType(BooleanManagedProperty) Then
                            Return DirectCast(x, BooleanManagedProperty).Value = DirectCast(y, BooleanManagedProperty).Value
                        ElseIf x.[GetType]() Is GetType(EntityReference) Then
                            If DirectCast(x, EntityReference).LogicalName = DirectCast(y, EntityReference).LogicalName Then
                                Return DirectCast(x, EntityReference).Id = DirectCast(y, EntityReference).Id
                            Else
                                Return False
                            End If
                        ElseIf x.GetType() Is GetType(Money) Then
                            Return (DirectCast(x, Money).Value = DirectCast(y, Money).Value)
                        ElseIf x.[GetType]() Is GetType(DateTime) OrElse x.[GetType]() Is GetType(System.Nullable(Of DateTime)) Then
                            'Compare only down to the second since CRM only tracks down to the second
                            Return Math.Abs((DirectCast(x, DateTime) - DirectCast(y, DateTime)).TotalSeconds) < 1
                        Else
                            Return x.Equals(y)
                        End If
                    Else
                        Return False
                    End If
                End If
            End Function
            Public Overloads Function GetHashCode(obj As Object) As Integer Implements IEqualityComparer.GetHashCode
                Return obj.GetHashCode()
            End Function
        End Class
        Public Enum eTextOptions
            ''' <summary>Ignore and let CRM handle any issues with the value</summary>
            Ignore
            ''' <summary>If the length is greater than the max length, truncate the value to the max length</summary>
            Truncate
            ''' <summary>Throw an error if the length of the value is greater than the max length</summary>
            ThrowError
        End Enum
        Public Enum eNumberOptions
            ''' <summary>Ignore and let CRM handle any issues with the value.</summary>
            Ignore
            ''' <summary>If the value is less than the min value set the value as the min value.<para>Let CRM handle any issues with the max value.</para></summary>
            CorrectMinIgnoreMax
            ''' <summary>If the value is less than the min value set the value as the min value.<para>Throw an error if the value is greater than the max value.</para></summary>
            CorrectMinThrowMax
            ''' <summary>If the value is greater than the max value set the value as the max value.<para>Let CRM handle any issues with the min value.</para></summary>
            CorrectMaxIgnoreMin
            ''' <summary>If the value is greater than the max value set the value as the max value.<para>Throw an error if the value is less than the min value.</para></summary>
            CorrectMaxThrowMin
            ''' <summary>If the value is less than the min value set the value as the min value.<para>If the value is greater than the max value set the value as the max value.</para></summary>
            CorrectMinAndMax
            ''' <summary>Throw an error if the value is less than the min or greater than the max</summary>
            ThrowError
        End Enum
        Public Enum eErrorType
            Text
            Number
        End Enum
    End Class

    Public Module ExtensionMethods
        <System.Runtime.CompilerServices.Extension> _
        Public Sub Create(Service As IOrganizationService, Proxy As BaseProxyClass)
            Dim EntityId As Guid = Service.Create(Proxy.Entity)
            Proxy.Id = EntityId
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub Update(Service As IOrganizationService, Proxy As BaseProxyClass)
            Proxy.Update(Service)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub Delete(Service As IOrganizationService, Proxy As BaseProxyClass)
            Service.Delete(Proxy.LogicalName, Proxy.Id)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub SetState(Service As IOrganizationService, Proxy As BaseProxyClass, State As Integer, Status As Integer)
            Service.SetState(Proxy, New OptionSetValue(State), New OptionSetValue(Status))
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub SetState(Service As IOrganizationService, Proxy As BaseProxyClass, State As OptionSetValue, Status As OptionSetValue)
            Dim Request As New SetStateRequest() With {.EntityMoniker = Proxy, .State = State, .Status = Status}
            Service.Execute(Request)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Function ToProxy(Of T As BaseProxyClass)(reference As EntityReference) As T
            If (reference Is Nothing) Then
                Return Nothing
            End If
            Return reference.ToEntity().ToProxy(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function ToEntity(reference As EntityReference) As Entity
            If (reference Is Nothing) Then
                Return Nothing
            End If
            Return New Entity() With { _
                .LogicalName = reference.LogicalName, _
                .Id = reference.Id _
            }
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function ToProxy(Of T As BaseProxyClass)(entity As Entity) As T
            If entity IsNot Nothing Then
                Dim Constructor As ConstructorInfo = GetType(T).GetConstructor(New Type() {GetType(Entity)})
                Return entity.ToProxy(Of T)(Constructor)
            End If
            Return Nothing
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Private Function ToProxy(Of T As BaseProxyClass)(entity As Entity, construcor As ConstructorInfo) As T
            Return TryCast(construcor.Invoke(New Object() {entity}), T)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetAttributeMetadata(Service As IOrganizationService, EntityLogicalName As String, AttributeLogicalName As String) As AttributeMetadata
            Dim Request As New RetrieveAttributeRequest() With { _
                .EntityLogicalName = EntityLogicalName, _
                .LogicalName = AttributeLogicalName _
            }
            Return TryCast(Service.Execute(Request), RetrieveAttributeResponse).AttributeMetadata
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetOptionMetadata(Value As OptionSetValue, Service As IOrganizationService, Entity As BaseProxyClass, AttributeLogicalName As String) As OptionMetadata
            Dim AttributeMeta As AttributeMetadata = Service.GetAttributeMetadata(Entity.LogicalName, AttributeLogicalName)
            If TypeOf AttributeMeta Is EnumAttributeMetadata Then
                Return DirectCast(AttributeMeta, EnumAttributeMetadata).GetOptionMetadata(Value.Value)
            Else
                Throw New Exception("The attribute is not an Enum type attribute")
            End If
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetOptionMetadata(EnumMeta As EnumAttributeMetadata, Value As Integer) As OptionMetadata
            Return (From Meta In EnumMeta.OptionSet.Options Where Meta.Value = Value Select Meta).FirstOrDefault()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetOptionSetText(EnumMeta As EnumAttributeMetadata, Value As Integer) As String
            Return EnumMeta.GetOptionMetadata(Value).GetOptionSetText()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetOptionSetText(Value As OptionSetValue, Service As IOrganizationService, Entity As BaseProxyClass, AttributeName As String) As String
            Dim OptionMeta As OptionMetadata = GetOptionMetadata(Value, Service, Entity, AttributeName)
            Return OptionMeta.GetOptionSetText()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetOptionSetText(OptionMeta As OptionMetadata) As String
            If OptionMeta IsNot Nothing AndAlso OptionMeta.Label IsNot Nothing AndAlso OptionMeta.Label.UserLocalizedLabel IsNot Nothing Then
                Return OptionMeta.Label.UserLocalizedLabel.Label
            End If
            Return String.Empty
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxies(Of T As BaseProxyClass)(Service As IOrganizationService, expression As ConditionExpression, columns As ColumnSet) As List(Of T)
            Dim FilterExpression As New FilterExpression()
            FilterExpression.AddCondition(expression)
            Return Service.RetrieveProxies(Of T)(FilterExpression, columns)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxies(Of T As BaseProxyClass)(Service As IOrganizationService, expression As FilterExpression, columns As ColumnSet) As List(Of T)
            Dim Query As New QueryExpression(BaseProxyClass.GetLogicalName(Of T)()) With { _
                .ColumnSet = columns, _
                .Criteria = expression _
            }
            Return Service.RetrieveProxies(Of T)(Query)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxies(Of T As BaseProxyClass)(Service As IOrganizationService, ParamArray Columns As String()) As List(Of T)
            Return Service.RetrieveProxies(Of T)(GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxies(Of T As BaseProxyClass)(Service As IOrganizationService, Columns As ColumnSet) As List(Of T)
            Dim Query As New QueryExpression(BaseProxyClass.GetLogicalName(Of T)()) With { _
                .ColumnSet = Columns _
            }
            Return Service.RetrieveMultipleAllPages(Query).ToProxies(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxies(Of T As BaseProxyClass)(Service As IOrganizationService, query As QueryExpression) As List(Of T)
            Dim Entities As EntityCollection = Service.RetrieveMultipleAllPages(query)
            Return Entities.ToProxies(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function ToProxies(Of T As BaseProxyClass)(Entities As EntityCollection) As List(Of T)
            Return Entities.Entities.ToProxies(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function ToProxies(Of T As BaseProxyClass)(Entities As IEnumerable(Of Entity)) As List(Of T)
            Dim Constructor As ConstructorInfo = GetType(T).GetConstructor(New Type() {GetType(Entity)})
            Return (From EntityTemp In Entities Select EntityTemp.ToProxy(Of T)(Constructor)).ToList()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxy(Of T As BaseProxyClass)(Entity As EntityReference, Service As IOrganizationService, ParamArray Columns As String()) As T
            Return Service.RetrieveProxy(Of T)(Entity.Id, Columns)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxy(Of T As BaseProxyClass)(Entity As EntityReference, Service As IOrganizationService, Columns As ColumnSet) As T
            Return Service.RetrieveProxy(Of T)(Entity.Id, Columns)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxy(Of T As BaseProxyClass)(Service As IOrganizationService, Id As Guid, ParamArray Columns As String()) As T
            Dim entityName As String = BaseProxyClass.GetLogicalName(Of T)()
            Return RetrieveEntity(New EntityReference(entityName, Id), Service, Columns).ToProxy(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveProxy(Of T As BaseProxyClass)(Service As IOrganizationService, Id As Guid, columns As ColumnSet) As T
            Dim entityName As String = BaseProxyClass.GetLogicalName(Of T)()
            Return RetrieveEntity(New EntityReference(entityName, Id), Service, columns).ToProxy(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxy(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, Value As Object, ParamArray Columns As String()) As T
            Return Service.FindProxy(Of T)(PropertyName, Value, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxy(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, Value As Object, Columns As ColumnSet) As T
            Return Service.FindProxies(Of T)(PropertyName, Value, Columns).FirstOrDefault()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxy(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, [Operator] As ConditionOperator, Value As Object, ParamArray Columns As String()) As T
            Return Service.FindProxy(Of T)(PropertyName, [Operator], Value, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxy(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, [Operator] As ConditionOperator, Value As Object, Columns As ColumnSet) As T
            Return Service.FindProxies(Of T)(PropertyName, [Operator], Value, Columns).FirstOrDefault()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxies(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, Value As Object, ParamArray Columns As String()) As List(Of T)
            Return Service.FindProxies(Of T)(PropertyName, Value, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxies(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, Value As Object, Columns As ColumnSet) As List(Of T)
            Return Service.FindProxies(Of T)(PropertyName, ConditionOperator.Equal, Value, Columns)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxies(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, [Operator] As ConditionOperator, Value As Object, ParamArray Columns As String()) As List(Of T)
            Return Service.FindProxies(Of T)(PropertyName, [Operator], Value, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function FindProxies(Of T As BaseProxyClass)(Service As IOrganizationService, PropertyName As String, [Operator] As ConditionOperator, Value As Object, Columns As ColumnSet) As List(Of T)
            Dim PropertyCondition As ConditionExpression = If(Value IsNot Nothing, New ConditionExpression(PropertyName, [Operator], Value), New ConditionExpression(PropertyName, [Operator]))
            Return Service.RetrieveProxies(Of T)(PropertyCondition, Columns)
        End Function
        Friend Function GetColumnSet(ParamArray Columns As String()) As ColumnSet
            Dim ColumnSet As New ColumnSet(Columns)
            If ColumnSet.Columns.Count = 0 Then
                ColumnSet = New ColumnSet(True)
            End If
            Return ColumnSet
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveEntity(Entity As EntityReference, Service As IOrganizationService, ParamArray Columns As String()) As Entity
            Return Entity.RetrieveEntity(Service, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveEntity(Entity As EntityReference, Service As IOrganizationService, Columns As ColumnSet) As Entity
            If Columns Is Nothing Then
                Columns = New ColumnSet(True)
            End If
            Return Service.Retrieve(Entity.LogicalName, Entity.Id, Columns)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveMultiple(Service As IOrganizationService, LogicalName As String, Condition As ConditionExpression, ParamArray Columns As String()) As EntityCollection
            Return Service.RetrieveMultiple(LogicalName, Condition, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveMultiple(Service As IOrganizationService, LogicalName As String, Condition As ConditionExpression, Columns As ColumnSet) As EntityCollection
            Dim Query As New QueryExpression(LogicalName) With { _
                .ColumnSet = Columns _
            }
            Query.Criteria.AddCondition(Condition)
            Return Service.RetrieveMultipleAllPages(Query)
        End Function
        Public RetrieveAllPages As Boolean = True
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveMultipleAllPages(Service As IOrganizationService, Query As QueryExpression) As EntityCollection
            Dim Results As EntityCollection = Service.RetrieveMultiple(Query)
            While Results.MoreRecords AndAlso RetrieveAllPages
                Dim CurrentResults As IList(Of Entity) = Results.Entities
                Query.PageInfo.PagingCookie = Results.PagingCookie
                Query.PageInfo.PageNumber += 1
                Results = Service.RetrieveMultiple(Query)
                Results.Entities.AddRange(CurrentResults)
            End While
            Return Results
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Sub RelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, Entities As IList(Of Entity))
            Service.RelateEntities(Target, IntermediateEntityName, Entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub RelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, ParamArray Entities As Entity())
            Dim References As IList(Of EntityReference) = (From EntityTemp In Entities Select EntityTemp.ToEntityReference()).ToList()
            Service.RelateEntities(Target, IntermediateEntityName, References)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub RelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, Entities As IList(Of EntityReference))
            Service.RelateEntities(Target, IntermediateEntityName, Entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub RelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, ParamArray Entities As EntityReference())
            Dim Request As New AssociateRequest() With { _
                .Relationship = New Relationship(IntermediateEntityName), _
                .Target = Target _
            }
            Request.RelatedEntities = New EntityReferenceCollection(Entities)
            Service.Execute(Request)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub UnRelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, Entities As IList(Of Entity))
            Service.UnRelateEntities(Target, IntermediateEntityName, Entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub UnRelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, ParamArray Entities As Entity())
            Dim References As IList(Of EntityReference) = (From EntityTemp In Entities Select EntityTemp.ToEntityReference()).ToList()
            Service.UnRelateEntities(Target, IntermediateEntityName, References)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub UnRelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, Entities As IList(Of EntityReference))
            Service.UnRelateEntities(Target, IntermediateEntityName, Entities.ToArray())
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub UnRelateEntities(Service As IOrganizationService, Target As EntityReference, IntermediateEntityName As String, ParamArray Entities As EntityReference())
            Dim Request As New DisassociateRequest() With { _
                .Target = Target, _
                .Relationship = New Relationship(IntermediateEntityName) _
            }
            Request.RelatedEntities = New EntityReferenceCollection(Entities)
            Service.Execute(Request)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub Add(Requests As List(Of OrganizationRequest), Proxy As BaseProxyClass)
            Dim RequestTemp As OrganizationRequest = GetOrganizationRequest(Proxy)
            If RequestTemp IsNot Nothing Then
                Requests.Add(RequestTemp)
            End If
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub Add(Collection As OrganizationRequestCollection, Proxy As BaseProxyClass)
            Dim RequestTemp As OrganizationRequest = GetOrganizationRequest(Proxy)
            If RequestTemp IsNot Nothing Then
                Collection.Add(RequestTemp)
            End If
        End Sub
        Public Function GetOrganizationRequest(Proxy As BaseProxyClass) As OrganizationRequest
            If Proxy.Entity.Id <> Guid.Empty AndAlso Proxy.IsDirty Then
                Return New UpdateRequest() With { _
                    .Target = Proxy.GetChangedEntity() _
                }
            ElseIf Proxy.Entity.Id = Guid.Empty Then
                Return New CreateRequest() With { _
                    .Target = Proxy _
                }
            End If
            Return Nothing
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Sub GrantAccess(Entity As IAccessControlledEntity, Service As IOrganizationService, Principal As EntityReference, AccessMask As AccessRights)
            Dim Request As New GrantAccessRequest() With { _
                .PrincipalAccess = New PrincipalAccess() With { _
                    .Principal = Principal, _
                    .AccessMask = AccessMask _
                }, _
                .Target = DirectCast(DirectCast(Entity, Object), EntityReference) _
            }
            Service.Execute(Request)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub ModifyAccess(Entity As IAccessControlledEntity, Service As IOrganizationService, Principal As EntityReference, AccessMask As AccessRights)
            Dim Request As New ModifyAccessRequest() With { _
                .PrincipalAccess = New PrincipalAccess() With { _
                    .Principal = Principal, _
                    .AccessMask = AccessMask _
                }, _
                .Target = DirectCast(DirectCast(Entity, Object), EntityReference) _
            }
            Service.Execute(Request)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub RevokeAccess(Entity As IAccessControlledEntity, Service As IOrganizationService, Revokee As EntityReference)
            Dim Request As New RevokeAccessRequest() With { _
                .Revokee = Revokee, _
                .Target = DirectCast(DirectCast(Entity, Object), EntityReference) _
            }
            Service.Execute(Request)
        End Sub
        Private mCachedEntities As New Dictionary(Of String, Dictionary(Of Guid, Entity))()
        Public Sub ClearCache(Optional EntityLogicalName As String = Nothing)
            SyncLock mCachedEntities
                If Not String.IsNullOrEmpty(EntityLogicalName) Then
                    If mCachedEntities.ContainsKey(EntityLogicalName) Then
                        mCachedEntities(EntityLogicalName) = New Dictionary(Of Guid, Entity)()
                    End If
                Else
                    mCachedEntities.Clear()
                End If
            End SyncLock
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub ClearCache(Service As IOrganizationService, Optional EntityLogicalName As String = Nothing)
            ClearCache(EntityLogicalName)
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveCachedEntityById(Service As IOrganizationService, EntityLogicalName As String, Id As Guid, Columns As ColumnSet) As Entity
            SyncLock mCachedEntities
                Dim CachedEntity As Entity = GetEntityFromCache(EntityLogicalName, Id)
                If CachedEntity Is Nothing Then
                    CachedEntity = Service.Retrieve(EntityLogicalName, Id, Columns)
                    AddEntityToCache(CachedEntity)
                End If
                Return CachedEntity
            End SyncLock
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetEntityFromCache(Service As IOrganizationService, EntityLogicalName As String, Id As Guid) As Entity
            Return GetEntityFromCache(EntityLogicalName, Id)
        End Function
        Public Function GetEntityFromCache(EntityLogicalName As String, Id As Guid) As Entity
            SyncLock mCachedEntities
                Dim CachedEntity As Entity = Nothing
                Dim CachedEntities As Dictionary(Of Guid, Entity) = GetEntitiesFromCache(EntityLogicalName)
                CachedEntities.TryGetValue(Id, CachedEntity)
                Return CachedEntity
            End SyncLock
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetEntitiesFromCache(Service As IOrganizationService, EntityLogicalName As String) As Dictionary(Of Guid, Entity)
            Return GetEntitiesFromCache(EntityLogicalName)
        End Function
        Public Function GetEntitiesFromCache(EntityLogicalName As String) As Dictionary(Of Guid, Entity)
            SyncLock mCachedEntities
                Dim CachedEntities As New Dictionary(Of Guid, Entity)()
                mCachedEntities.TryGetValue(EntityLogicalName, CachedEntities)
                Return If(CachedEntities IsNot Nothing, CachedEntities, New Dictionary(Of Guid, Entity)())
            End SyncLock
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Sub AddEntitiesToCache(Of T As BaseProxyClass)(Service As IOrganizationService, Entities As ICollection(Of T))
            For Each Item As T In Entities
                AddEntityToCache(Item.Entity)
            Next
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub AddEntitiesToCache(Service As IOrganizationService, Entities As ICollection(Of Entity))
            AddEntitiesToCache(Entities)
        End Sub
        Public Sub AddEntitiesToCache(Entities As ICollection(Of Entity))
            SyncLock mCachedEntities
                For Each Item As Entity In Entities
                    AddEntityToCache(Item)
                Next
            End SyncLock
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Sub AddEntityToCache(Service As IOrganizationService, Item As Entity)
            AddEntityToCache(Item)
        End Sub
        Public Sub AddEntityToCache(Item As Entity)
            SyncLock mCachedEntities
                If Not mCachedEntities.ContainsKey(Item.LogicalName) Then
                    mCachedEntities.Add(Item.LogicalName, New Dictionary(Of Guid, Entity)() From { _
                    })
                End If
                If Not mCachedEntities(Item.LogicalName).ContainsKey(Item.Id) Then
                    mCachedEntities(Item.LogicalName).Add(Item.Id, Item)
                End If
            End SyncLock
        End Sub
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveCachedProxy(Of T As BaseProxyClass)(Service As IOrganizationService, Entity As EntityReference) As T
            Return RetrieveCachedProxy(Of T)(Service, Entity, New ColumnSet(True))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveCachedProxy(Of T As BaseProxyClass)(Service As IOrganizationService, Entity As EntityReference, ParamArray Columns As String()) As T
            Return RetrieveCachedProxy(Of T)(Service, Entity, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveCachedProxy(Of T As BaseProxyClass)(Service As IOrganizationService, Entity As EntityReference, Columns As ColumnSet) As T
            Return RetrieveCachedEntityById(Service, Entity.LogicalName, Entity.Id, Columns).ToProxy(Of T)()
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveCachedProxy(Of T As BaseProxyClass)(Entity As EntityReference, Service As IOrganizationService) As T
            Return RetrieveCachedProxy(Of T)(Entity, Service, New ColumnSet(True))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveCachedProxy(Of T As BaseProxyClass)(Entity As EntityReference, Service As IOrganizationService, ParamArray Columns As String()) As T
            Return RetrieveCachedProxy(Of T)(Entity, Service, GetColumnSet(Columns))
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function RetrieveCachedProxy(Of T As BaseProxyClass)(Entity As EntityReference, Service As IOrganizationService, Columns As ColumnSet) As T
            Return Service.RetrieveCachedProxy(Of T)(Entity, Columns)
        End Function
        <System.Runtime.CompilerServices.Extension> _
        Public Function GetEntityMetadata(Service As IOrganizationService, LogicalName As String, Filters As EntityFilters) As EntityMetadata
            Dim Request As New RetrieveEntityRequest() With { _
                .LogicalName = LogicalName, _
                .EntityFilters = Filters _
            }
            Return DirectCast(Service.Execute(Request), RetrieveEntityResponse).EntityMetadata
        End Function
    End Module
    Public Interface IAccessControlledEntity
    End Interface
End Namespace